#!/usr/bin/env bash

# GENERIC B5 FUNCTIONS

B5_FONT_GREEN='\033[00;32m'
B5_FONT_RED='\033[00;31m'
B5_FONT_RESTORE='\033[0m'

b5:init() {
    local directory="${1:-}"
    local template="${2:-}"

    if [ -z "${directory}" ]
    then
        echo "No directory given. Abort task run."
        echo "Usage: b5_init <directory> [template]"
        return 1
    fi

    if [ -z "${template}" ]
    then
        template='minimal'
    fi

    if [ -d "${directory}" ]
    then
        echo "Directory already exists. Aborting!"
        return 1
    fi

    if [ ! -d "${TEMPLATES_PATH}/${template}" ]
    then
        echo "Template does not exist. Aborting!"
        return 1
    fi

    (
        mkdir -p "${directory}" && \
        shopt -s dotglob
        cp -r "${TEMPLATES_PATH}/${template}/src/"* "${directory}"/ && \
        shopt -u dotglob
        cd "${directory}" && \
        git init . && (
            if [ -e "${TEMPLATES_PATH}/${template}/init.sh" ]
            then
                source "${TEMPLATES_PATH}/${template}/init.sh" || b5:abort "Initialization of new project (${directory}) failed"
            fi
        ) && \
        echo "All set up. You may now start using ${directory}"
    )
}

b5:abort() {
    echo -e "${B5_FONT_RED}${1:-}${B5_FONT_RESTORE}"
    exit 1
}

b5:error_exit() {
    local err=$?
    set +o xtrace
    local code="${1:-1}"
    echo
    echo "Error in ${BASH_SOURCE[1]}:${BASH_LINENO[0]}. '${BASH_COMMAND}' exited with status $err"
    # Print out the stack trace described by $function_stack
    if [ ${#FUNCNAME[@]} -gt 2 ]
    then
        echo "Call tree:"
        for ((i=1;i<${#FUNCNAME[@]}-1;i++))
        do
            echo " $i: ${BASH_SOURCE[$i+1]}:${BASH_LINENO[$i]} ${FUNCNAME[$i]}(...)"
        done
    fi
    echo "Exiting with status ${code}"
    echo -e "\n${B5_FONT_RED}Task failed, see error above${B5_FONT_RESTORE}"
    exit "${code}"
}

b5:run() {
    "$@"
    echo -e "\n${B5_FONT_GREEN}Task exited ok${B5_FONT_RESTORE}"
}

b5:help() {
    echo "$0 <task> <args>"
    echo "Tasks:"
    compgen -A function | sed -En 's/task:(.*)/\1/p' | cat -n
#    echo
#    echo "Debug for development:"
#    b5:config_dump
}

b5:install() {
    if [ -e "local.example.yml" ] && [ ! -e "local.yml" ]
    then
        cp "local.example.yml" "local.yml"
        echo
        echo -e "${B5_FONT_GREEN}IMPORTANT IMPORTANT IMPORTANT IMPORTANT IMPORTANT IMPORTANT IMPORTANT${B5_FONT_RESTORE}"
        echo
        echo "Copied local.example.yml to local.yml"
        echo "Make sure to edit local.yml now"
        echo
        echo -e "${B5_FONT_GREEN}IMPORTANT IMPORTANT IMPORTANT IMPORTANT IMPORTANT IMPORTANT IMPORTANT${B5_FONT_RESTORE}"
        echo
    fi
}

# YAML (See https://github.com/jasperes/bash-yaml)

b5:yaml_parse() {
    if [ -z "${1:-}" ]
    then
        echo "Usage: b5:yaml_parse <filename> [prefix]"
        return 1
    fi
    local yaml_file="$1"
    local prefix="${2:-}"
    local s
    local w
    local fs

    s='[[:space:]]*'
    w='[a-zA-Z0-9_]*'
    fs="$(echo @|tr @ '\034')"

    (
        sed -ne 's/--//g; s/\"/\\\"/g; s/\#.*//g; s/\s*$//g;' \
            -e  "s|^\($s\)\($w\)$s:$s\"\(.*\)\"$s\$|\1$fs\2$fs\3|p" \
            -e "s|^\($s\)\($w\)$s[:-]$s\(.*\)$s\$|\1$fs\2$fs\3|p" |
        awk -F"$fs" '{
            indent = length($1)/2;
            if (length($2) == 0) { conj[indent]="+";} else {conj[indent]="";}
            vname[indent] = $2;
            for (i in vname) {if (i > indent) {delete vname[i]}}
                if (length($3) > 0) {
                    vn=""; for (i=0; i<indent; i++) {vn=(vn)(vname[i])("_")}
                    printf("%s%s%s%s=(\"%s\")\n", "'"$prefix"'",vn, $2, conj[indent-1],$3);
                }
            }' |
        sed 's/_=/+=/g'
    ) < "$yaml_file"
}

b5:yaml_load() {
    if [ -z "${1:-}" ]
    then
        echo "Usage: b5:yaml_parse <filename> [prefix]"
        return 1
    fi
    local yaml_file="$1"
    local yaml_prefix="${2:-}"
    eval "$( b5:yaml_parse "$yaml_file" "$yaml_prefix" )"
}

# CONFIG

b5:config_load() {
    if [ -e "config.yml" ]
    then
        b5:yaml_load "config.yml" "CONFIG_"
    fi
    if [ -e "local.yml" ]
    then
        b5:yaml_load "local.yml" "CONFIG_local_"
    fi
}

b5:config_dump() {
    if [ -e "config.yml" ]
    then
        b5:yaml_parse "config.yml" "CONFIG_"
    fi
    if [ -e "local.yml" ]
    then
        b5:yaml_parse "local.yml" "CONFIG_local_"
    fi
}

b5:config_get() {
    if [ -z "${1:-}" ]
    then
        echo "Usage: config_get <variable_name> [default]"
        exit 1
    fi

    local config_name="CONFIG_$1"
    local default="$2"
    eval "echo \"\${${config_name}:-${default}}\""
}

# MODULES

declare -a MODULES_LOADED
MODULES_LOADED=("b5")

b5:module_exists() {
    if [ -z "${1:-}" ]
    then
        echo "Usage: b5:module_exists <module_name>"
        return 254
    fi
    if [ -e "${MODULES_PATH}/${1}.sh" ]
    then
        return 0
    else
        return 1
    fi
}

b5:module_load() {
    if [ -z "${1:-}" ]
    then
        echo "Usage: b5:module_load <module_name>"
        return 254
    fi
    if [[ " ${MODULES_LOADED[*]} " == *" $1 "* ]]
    then
        return 0
    fi
    if ( b5:module_exists "$1" )
    then
        source "${MODULES_PATH}/${1}.sh" && \
        MODULES_LOADED+=("$1")
    else
        echo -e "${B5_FONT_RED}Module $1 not found${B5_FONT_RESTORE}"
        echo -e "${B5_FONT_RED}ABORTING!${B5_FONT_RESTORE}"
        exit 250
    fi
}

b5:module_has() {
    local module="${1:-}"
    local action="${2:-}"
    if [ -z "$action" ] || [ -z "$module" ]
    then
        echo "Usage: b5:module_run <module> <module_action>"
        return 254
    fi

    local type_result=$( type -t "${module}:${action}" )
    if [ "$type_result" == "function" ]
    then
        return 0
    else
        return 1
    fi
}

b5:module_run() {
    local module="${1:-}"
    local action="${2:-}"
    if [ -z "$action" ] || [ -z "$module" ]
    then
        echo "Usage: b5:module_run <module> <module_action>"
        return 254
    fi
    shift 2

    echo "Calling ${module}:${action}" && \
        "${module}:${action}" "$@" && \
        echo
}

b5:module_run_all() {
    local action="${1:-}"
    if [ -z "$action" ]
    then
        echo "Usage: b5:module_run_all <module_action>"
        return 254
    fi
    shift

    for module in ${MODULES_LOADED[@]}
    do
        if b5:module_has "${module}" "${action}"
        then
            b5:module_run "${module}" "${action}" "$@"
        fi
    done
}
